import { GraphType, LinkType, NodeType } from '../Types';
import { tailwindColors } from 'config';
import { useEffect, useMemo, useRef, useState } from 'react';
import { Application, Circle, Container, FederatedPointerEvent, Graphics, Point, Sprite, Texture, autoDetectRenderer } from 'pixi.js';
import { binaryColor, nodeColor as nodeColor } from './utils';
import { select, zoom as d3zoom, drag as d3drag } from 'd3';
import * as force from './NLForce';
import { Viewport } from 'pixi-viewport';
import ResultNodeLinkParserUseCase from '../../shared/ResultNodeLinkParserUseCase';
import { GraphQueryResult, GraphQueryResultFromBackendPayload } from '@graphpolaris/shared/lib/data-access/store/graphQueryResultSlice';

type Props = {
  windowSize: { width: number; height: number };
  onClick: (node: NodeType) => void;
  highlightNodes: NodeType[];
  currentShortestPathEdges?: LinkType[];
  highlightedLinks?: LinkType[];
  jaccardThreshold?: number;
  myDiv: HTMLDivElement | null;
};

const app = new Application({ background: 0xffffff, antialias: true, autoDensity: true, eventMode: 'auto' });
const nodes = new Container();
const links = new Container();

//////////////////
// MAIN COMPONENT
//////////////////

export const useNLPixi = (props: Props) => {
  const [drag, setDrag] = useState({ x: 0, y: 0 });
  const nodeMap = useRef(new Map<string, Graphics>());
  const linkMap = useRef(new Map<string, Graphics>());
  const graph = useRef<GraphType>();
  const viewport = useRef<Viewport>();

  useEffect(() => {
    app.renderer.resize(props.windowSize.width, props.windowSize.height);
    app.render();
  }, [props.windowSize]);

  useEffect(() => {
    if (props.myDiv && props.myDiv.children.length === 0) props.myDiv.appendChild(app.view as HTMLCanvasElement);
  }, [props.myDiv]);

  const updateNode = (node: NodeType) => {
    // console.log(item);
    const gfx = nodeMap.current.get(node.id);
    gfx?.position.set(node.x, node.y);

    // if (!item.position) {
    //   item.position = new Point(node.x, node.y);
    // } else {
    //   item.position.set(node.x, node.y);
    // }
    // Update attributes position if they exist
    // if (node.gfxAttributes) {
    //   const x = node.x - node.gfxAttributes.width / 2;
    //   const y = node.y - node.gfxAttributes.height - 20;
    //   if (!node.gfxAttributes?.position) node.gfxAttributes.position = new Point(x, y);
    //   else {
    //     node.gfxAttributes.position.set(x, y);
    //   }
    // }
  };

  const dragging = useRef<{ node: NodeType; gfx: Graphics } | null>(null);
  function onDragStart(event: FederatedPointerEvent, node: NodeType, gfx: Graphics) {
    // store a reference to the data
    // the reason for this is because of multitouch
    // we want to track the movement of this particular touch
    event.stopPropagation();
    if (viewport.current) viewport.current.pause = true;
    dragging.current = { node, gfx };
  }

  function onDragEnd(event: FederatedPointerEvent) {
    if (dragging.current) {
      event.stopPropagation();
      dragging.current.node.fx = null;
      dragging.current.node.fy = null;
      dragging.current = null;
      if (viewport.current) viewport.current.pause = false;
    }
  }

  function onDragMove(event: FederatedPointerEvent) {
    if (dragging.current) {
      event.stopPropagation();
      console.log(viewport.current?.scaled);

      if (!dragging.current.node.fx) dragging.current.node.fx = dragging.current.node.x || 0;
      if (!dragging.current.node.fy) dragging.current.node.fy = dragging.current.node.y || 0;
      dragging.current.node.fx += event.movementX / (viewport.current?.scaled || 1);
      dragging.current.node.fy += event.movementY / (viewport.current?.scaled || 1);
      force.simulation.alpha(0.1).restart();
    }
  }

  const createNode = (node: NodeType, selected?: boolean) => {
    // check if node is already drawn, and if so, delete it
    if (node && node?.id && nodeMap.current?.has(node.id)) {
      nodeMap.current.delete(node.id);
    }
    const gfx = new Graphics();
    node.selected = selected;
    const lineColor = selected ? tailwindColors.entity[400] : '#000000';

    const lineWidth = selected ? 3 : 1.5;
    gfx.lineStyle(lineWidth, binaryColor(lineColor));
    //check if not undefined.
    if (node.cluster) {
      gfx.beginFill(nodeColor(node.cluster));
    } else {
      //if cluster is undefined, use type to determine the color.
      //check if not undefined.
      if (node.type) {
        gfx.beginFill(nodeColor(node.type));
      }
    }
    gfx.drawCircle(0, 0, Math.max(node.radius || 5, 5));
    gfx.hitArea = new Circle(0, 0, 4);
    gfx.name = 'node_' + node.id;
    gfx.position.set(node.x, node.y);
    gfx.eventMode = 'dynamic';
    gfx.on('mousedown', (e) => onDragStart(e, node, gfx));
    gfx.on('mousemove', onDragMove);
    gfx.on('mouseup', onDragEnd);

    nodes.addChild(gfx);
    nodeMap.current.set(node.id, gfx);
    return gfx;
  };

  /** UpdateRadius works just like UpdateColors, but also applies radius*/
  const UpdateRadius = (graph: GraphType, radius: number) => {
    // update for each node in graph
    graph.nodes.forEach((node: NodeType) => {
      createNode(node);
    });
  };

  const updateLink = (link: LinkType) => {
    if (!graph.current) return;

    const _source = link.source;
    const _target = link.target;

    if (!_source || !_target) {
      console.log('source or target not found', _source, _target);
      return;
    }

    let sourceId = '';
    let targetId = '';
    let source: NodeType | undefined;
    let target: NodeType | undefined;

    if (typeof _source === 'string') {
      sourceId = link.source as string;
      targetId = link.target as string;
      source = nodeMap.current.get(sourceId) as NodeType | undefined;
      target = nodeMap.current.get(targetId) as NodeType | undefined;
    } else {
      source = link.source as NodeType;
      target = link.target as NodeType;
      sourceId = source.id;
      targetId = target.id;
    }
    if (!source || !target) {
      console.log('source or target not found', source, target, sourceId, targetId);
      return;
    }

    const id = sourceId + targetId;
    const gfx = linkMap.current.get(id);

    if (gfx) {
      gfx.clear();
      gfx.beginFill();
      // Check if link is a machine learning edge
      //   if (link.mlEdge && props.jaccardThreshold) {
      //     if (link.value > props.jaccardThreshold) {
      //       gfx.moveTo(source.x, source.y);
      //       gfx.lineTo(target.x, target.y);
      //       gfx.lineStyle(link.value * 1.8, 0x0000ff);
      //     }
      //   } else gfx.lineStyle(0.2, 0x000000);
      //   //Check if link is hightlighted
      //   if (props.highlightedLinks && props.highlightedLinks.includes(link)) {
      //     //If highlighted and ML edge make it purple
      //     if (link.mlEdge && props.jaccardThreshold) {
      //       if (link.value > props.jaccardThreshold) {
      //         gfx.lineStyle(link.value * 1.8, 0xaa00ff);
      //       }
      //       //If highlight and not ML edge make it red
      //     } else {
      //       gfx.lineStyle(1.0, 0xff0000);
      //     }
      //   }
      //   if (props.currentShortestPathEdges && props.currentShortestPathEdges.includes(link)) {
      //     gfx.lineStyle(3.0, 0x00ff00);
      //   }
      //   console.log('link', source.x);

      //   gfx.position.set(source.x, source.y);
      //   gfx.moveTo(source.x || 0, source.y || 0);
      //   gfx.lineTo(target.x || 1000, target.y || 1000);
      //   gfx.position.set(source.x, source.y);
      gfx
        .lineStyle(0.5, 0x000000)
        .moveTo(source.x || 0, source.y || 0)
        .lineTo(target.x || 0, target.y || 0);
      gfx.endFill();
    }
  };

  const createLink = (link: LinkType) => {
    const sourceId = link.source as string;
    const targetId = link.target as string;
    const id = sourceId + targetId;

    if (linkMap.current.has(id)) {
      linkMap.current.delete(id);
    }
    const gfx = new Graphics();
    gfx.name = 'link_' + id;
    linkMap.current.set(id, gfx);
    updateLink(link);
    links.addChild(gfx);
    return gfx;
  };

  /**
   * SetNodeGraphics is an initializing function. It attaches the nodes and links to the simulation.
   * It creates graphic objects and adds these to the PIXI containers. It also clears both of these of previous nodes and links.
   * @param graph The graph returned from the database and that is parsed into a nodelist and edgelist.
   * @param radius The radius of the node.
   */
  const refresh = (graphQueryResult: GraphQueryResult) => {
    const resultNodeLinkParserUseCase = new ResultNodeLinkParserUseCase();
    graph.current = resultNodeLinkParserUseCase.parseQueryResult(graphQueryResult);

    nodes.removeChildren();
    links.removeChildren();
    app.stage.removeChildren();

    viewport.current = new Viewport({
      screenWidth: props.windowSize.width,
      screenHeight: props.windowSize.height,
      worldWidth: props.windowSize.width,
      worldHeight: props.windowSize.height,
      stopPropagation: true,
      events: app.renderer.events, // the interaction module is important for wheel to work properly when renderer.view is placed or scaled
    });
    app.stage.addChild(viewport.current);
    // activate plugins
    viewport.current.drag().pinch().wheel().animate({}).decelerate({ friction: 0.75 });

    viewport.current.addChild(links);
    viewport.current.addChild(nodes);
    // app.stage.addChild(links);
    // app.stage.addChild(nodes);
    app.stage.eventMode = 'dynamic';
    app.stage.on('mouseup', onDragEnd);
    app.stage.on('pointerup', onDragEnd);

    app.ticker.add(function (delta) {
      if (graph.current) {
        graph.current.nodes.forEach((node: any) => {
          const gfx = nodeMap.current.get(node.id);
          if (!gfx) return;
          const pos = gfx.position;
          //   gfx.position.set(pos.x + (node.x - pos.x) / delta, pos.y + (node.y - pos.y) / delta);
          gfx.position.set(node.x, node.y);
        });

        // Update forces of the links
        graph.current.links.forEach((link: any) => {
          updateLink(link);
        });
      }
    });

    // app.stage.onmousedown = (e) => {
    //   isPanning.current = true;
    //   console.log('mousedown');
    // };
    // app.stage.onmousemove = (e) => {
    //   if (isPanning.current) {
    //     setPan({ x: pan.x + e.movementX, y: pan.y + e.movementY });
    //     app.stage.setTransform(pan.x + e.movementX, pan.y + e.movementY, scale, scale);
    //   }
    // };
    // app.stage.onmouseup = (e) => {
    //   isPanning.current = false;
    //   console.log('mouseup');
    // };

    // Create and initialise graphic objects for the nodes
    if (graph && graph.current.nodes && graph.current.links) {
      nodeMap.current.clear();
      graph.current.nodes.forEach((node: NodeType) => {
        createNode(node);
      });
      linkMap.current.clear();
      graph.current.links.forEach((link: LinkType) => {
        createLink(link);
      });

      // // update text colour (written after nodes so that text appears on top of nodes)
      //   nodes.forEach((node: NodeType) => {
      //   if (node.gfxAttributes !== undefined) {
      //       const selected = node.selected === true;
      //       node.gfxAttributes.destroy();
      //       createAttributes(node);
      //       if (selected) {
      //       showAttributes(node);
      //       }
      //   }
      //   });

      // // refresh
      //   force.simulation.alphaTarget(0).restart();

      force.startSimulation(graph.current, props.windowSize);
      force.simulation.on('tick', () => {});
    }
  };

  return { renderer: app.renderer, stage: app.stage, links, refresh, graph: graph.current };
};
